//
//  WaitingForPlayersScene.m
//  PeerPong
//
//  Created by Volker Tietz on 11/23/11.
//  Copyright (c) 2011 __MyCompanyName__. All rights reserved.
//

#import "MainScene.h"
#import "BackgroundLayer.h"
#import "WelcomeLayer.h"
#import "PaddlesLayer.h"
#import "BallLayer.h"
#import "GameBoxLayer.h"
#import "PeerPongGameplay.h"
#import "PeerPongPlayer.h"
#import "QueueLayer.h"
#import "MessageLayer.h"
#import "MyContactListener.h"
#import "SimpleAudioEngine.h"

@implementation MainScene

-(id)init {
    self = [super init];
    if (self) {
        gameplay = [PeerPongGameplay sharedInstance];
        soundDelayFrames = 0;
        
        backgroundLayer = [BackgroundLayer node];
        [self addChild:backgroundLayer];
        
        gameBoxLayer = [GameBoxLayer node];
        [self addChild:gameBoxLayer];
        
        queueLayer = [QueueLayer node];
        [self addChild:queueLayer];
        
        [self addWelcomeLayer];
        
        paddlesLayer = [PaddlesLayer node];
        [self addChild:paddlesLayer];
        
        contactListener = new MyContactListener();
        gameplay.world->SetContactListener(contactListener);
        
        messageLayer = [MessageLayer node];
        [self addChild:messageLayer z:100];
        
        // Register tick method which is called on every frame rendering
        [self schedule:@selector(tick:)];
        
        [[SimpleAudioEngine sharedEngine] playBackgroundMusic:@"peer_pong_intro_128k.mp3"];
    }
    
    return self;
}

#pragma mark - Manage paddles

-(void)updatePaddles {
    [paddlesLayer updatePaddles];
}

-(void)addPaddleForPlayer:(PeerPongPlayer *)aPlayer {
    [gameBoxLayer removeEdgeAndBarForPlayer:aPlayer];
    [paddlesLayer addPaddleForPlayer:aPlayer];
}

-(void)removePaddleForPlayer:(PeerPongPlayer *)aPlayer {
    [gameBoxLayer addEdgeAndBarForSlotNumber:[aPlayer.slot intValue]];
    [paddlesLayer removePaddleForPlayer:aPlayer];
    
    [self lastPlayerLeavesGame];
}

#pragma mark - Manage layers

-(void)addBallLayer {
    [[SimpleAudioEngine sharedEngine] stopBackgroundMusic];
    [self removeWelcomeLayer];
    ballLayer = [BallLayer node];
    [self addChild:ballLayer z:9 tag:98];
}

-(void)removeBallLayer {
    [ballLayer removeBall];
    [self removeChild:ballLayer cleanup:YES];
}

-(void)addWelcomeLayer {
    welcomeLayer = [WelcomeLayer node];
    [self addChild:welcomeLayer z:10 tag:99];
}

-(void)removeWelcomeLayer {
    if([self getChildByTag:99]) {
        [self removeChildByTag:99 cleanup:YES];
    }
}

-(void)lastPlayerLeavesGame {
    // Check if last player leaves game
    if(gameplay.gameIsRunning == YES && [gameplay.players count] == 1) {
        [self endGame];
    }
}

-(void)reinitGame {
    if(gameplay.gameIsRunning == YES) {
        [self removeBallLayer];
        [self addBallLayer];
    }
}

-(void)endGame {
    gameplay.gameIsRunning = NO;
    [self removeBallLayer];
    [self addWelcomeLayer];
    [[SimpleAudioEngine sharedEngine] playBackgroundMusic:@"peer_pong_intro_128k.mp3"];
}

#pragma mark - Manage queue items

-(void)updateQueue {
    [queueLayer update];
}

#pragma mark - Pop Messages
-(void)popMessage:(NSString*)aMessage{
    [messageLayer popMessage:aMessage];
}

#pragma mark - Tick

- (void)tick:(ccTime) dt {
    @synchronized(self) {
        gameplay.world->Step(dt, 10, 10);    
        for(b2Body *b = gameplay.world->GetBodyList(); b; b=b->GetNext()) {    
            if (b->GetUserData() != NULL) {
                CCSprite *sprite = (CCSprite *)b->GetUserData(); 
                            
                CGPoint position = ccp(b->GetPosition().x * gameplay.PTM_RATIO,
                                       b->GetPosition().y * gameplay.PTM_RATIO);
                
                // Set new calculated position for sprite
                sprite.position = position;
                sprite.rotation = -1 * CC_RADIANS_TO_DEGREES(b->GetAngle());
                    
                // Ball sprite specific 
                if(sprite.tag == gameplay.ballTag) {
                    [gameplay ballIsOnPosition:position];
                    
                    static int maxSpeed = 30;
                    static int minSpeed = 20;
                    
                    b2Vec2 velocity = b->GetLinearVelocity();
                    float32 speed = velocity.Length();
                    
                    // damp if max speed exceeded
                    if (speed > maxSpeed) {
                        b->SetLinearDamping(0.5);
                    } else if (speed < maxSpeed) {
                        b->SetLinearDamping(0.0);
                    }
                    
                    if (speed < minSpeed) {
                        b->SetLinearDamping(-0.5);
                    } else if (speed > minSpeed) {
                        b->SetLinearDamping(0.0);
                    }
                }
                
                
            }        
        }
        
        if(soundDelayFrames > 0) soundDelayFrames--;
        
        std::vector<MyContact>::iterator pos;
        BOOL hitsWall = NO;
        for(pos = contactListener->_contacts.begin(); pos != contactListener->_contacts.end(); ++pos) {
            MyContact contact = *pos;
            
            if(soundDelayFrames <= 0) {
                
                // 
                if ((contact.fixtureA == ballLayer.fixture && contact.fixtureB == gameBoxLayer.leftEdge) ||
                    (contact.fixtureA == gameBoxLayer.leftEdge && contact.fixtureB == ballLayer.fixture)) {
                    hitsWall = YES;
                }
                
                if ((contact.fixtureA == ballLayer.fixture && contact.fixtureB == gameBoxLayer.topEdge) ||
                    (contact.fixtureA == gameBoxLayer.topEdge && contact.fixtureB == ballLayer.fixture)) {
                    hitsWall = YES;
                }
                
                if ((contact.fixtureA == ballLayer.fixture && contact.fixtureB == gameBoxLayer.rightEdge) ||
                    (contact.fixtureA == gameBoxLayer.rightEdge && contact.fixtureB == ballLayer.fixture)) {
                    hitsWall = YES;
                }
                
                if ((contact.fixtureA == ballLayer.fixture && contact.fixtureB == gameBoxLayer.bottomEdge) ||
                    (contact.fixtureA == gameBoxLayer.bottomEdge && contact.fixtureB == ballLayer.fixture)) {
                    hitsWall = YES;
                }
                if(hitsWall) [gameplay playSound:kSoundBallOnWall];
                
                soundDelayFrames = 10;
            }
            
            

        }
    }
}

- (void)dealloc {
    delete contactListener;
    [super dealloc];
}

@end
