//
//  PeerPongGameplay.m
//  PeerPong
//
//  Created by Volker Tietz on 11/23/11.
//  Copyright (c) 2011 __MyCompanyName__. All rights reserved.
//

#import "PeerPongGameplay.h"
#import "PeerPongServer.h"
#import "MainScene.h"
#import "WelcomeLayer.h"
#import "UnicornServer.h"
#import "PeerPongPlayer.h"
#import "SimpleAudioEngine.h"
#import "GameOver.h"
#import "LostLife.h"

@implementation PeerPongGameplay

@synthesize peerPongServer, serverName, world, groundBody, windowSize, 
            players, playerQueue, PTM_RATIO, GROUND_BOX_PADDING, gameIsRunning, 
            fontName, defaultFontSize, ballTag, queueSpriteTag;


+(PeerPongGameplay *)sharedInstance {
    static dispatch_once_t pred = 0;
    __strong static id _sharedObject = nil;
    dispatch_once(&pred, ^{
        _sharedObject = [[self alloc] init];
    });
    return _sharedObject;
}

- (id)init {
    self = [super init];
    if (self) {
        self.GROUND_BOX_PADDING = 50.0;
        self.PTM_RATIO = 32.0;

        self.fontName = @"Helvetica-Bold";
        self.defaultFontSize = 48;
        
        self.ballTag = 50;
        self.queueSpriteTag = 60;
        
        playerLock = [NSLock new];
    }
    return self;
}

#pragma mark - Server callbacks

-(void)receivedCommand:(Command *)command fromClient:(GCDAsyncSocket *) client {
    PeerPongPlayer *player = [self getPlayerSlotBySocket:client];
    if (player != nil) {
        player.offset = [command.offset floatValue];
        
        MainScene *scene = (MainScene *) [[CCDirector sharedDirector] runningScene];
        [scene updatePaddles];

        
    } else {
        NSLog(@"Received Command: %@ from unknown player", command);
    }
}

-(void)receivedHello:(Hello *)hello fromClient:(GCDAsyncSocket *)client {
    @synchronized(playerLock) {
        NSLog(@"Received hello %@ from client %@", hello, client);
        PeerPongPlayer *player = [self addPlayerWithHello:hello andSocket:client];
        
        // TODO: Logging falls player zweimal hello gesendet hat
        if(player != nil) {
            currentScene = [[CCDirector sharedDirector] runningScene];
            
            MainScene *scene = (MainScene *) [[CCDirector sharedDirector] runningScene];
            dispatch_async(dispatch_get_main_queue(), ^{
                [scene updateQueue];
                [self askFirstPlayerForReadiness];
            });

        }
    }
}

-(void)receivedIAmReadyForSlot:(NSNumber*)aSlot fromClient:(GCDAsyncSocket *)client {
    NSLog(@"Received IAmReady from client %@ for slot %d", client, [aSlot intValue]);
    PeerPongPlayer *player = [self movePlayerFromQueueToSlot:aSlot WithSocket:client];
    MainScene *scene = (MainScene *) [[CCDirector sharedDirector] runningScene];
    dispatch_async(dispatch_get_main_queue(), ^{
        [scene addPaddleForPlayer:player];
        [scene updateQueue];
        if(self.gameIsRunning == NO) {
            [self startGame];
        }
    });
    [self askFirstPlayerForReadiness];
}

-(void)receivedDisconnectFromClient:(GCDAsyncSocket *)client {
    NSLog(@"Player %d disconnected!", [[self getPlayerSlotBySocket:client].slot intValue]);
    dispatch_async(dispatch_get_main_queue(), ^{
        [self removePlayerBySocket:client];
    });
}

#pragma mark - Game controls

-(void)initiateGame {
    [self startServer];
    self.gameIsRunning = NO;
    
    // start scene if not already running
    if(![[CCDirector sharedDirector] runningScene]){
        [[CCDirector sharedDirector] runWithScene: [MainScene node]];
    } else {
        [[CCDirector sharedDirector] replaceScene: [MainScene node]];
    }
    
}

-(void)destroyGame{
    [self stopServer];
    self.gameIsRunning = NO;
}

-(void)startGame {
    self.gameIsRunning = YES;
    MainScene *scene = (MainScene *)[[CCDirector sharedDirector] runningScene];
    [scene addBallLayer];
}

-(void)startServer {
    // reset server
    delete self.world;
    self.world = nil;
    self.peerPongServer = nil;
    self.players = nil;
    self.playerQueue = nil;
    
    // setup world
    b2Vec2 gravity = b2Vec2(0.0f, 0.0f);
    bool doSleep = true;
    self.world = new b2World(gravity, doSleep);
    self.windowSize = [[CCDirector sharedDirector] winSize];
    self.players = [[[NSMutableDictionary alloc] init] autorelease];
    self.playerQueue = [[[NSMutableArray alloc] init] autorelease];
    
    // start server
    self.peerPongServer = [[[PeerPongServer alloc] init] autorelease];
    [self.peerPongServer start]; 
    self.serverName = self.peerPongServer.server.name;
    self.peerPongServer.gamePlay = self;
}

-(void)stopServer {
    [self.peerPongServer stop];
}

#pragma mark - Sound

-(void)playSound:(int)aSound {
    return;
    switch (aSound) {
        case kSoundPlayerJoined:
            [[SimpleAudioEngine sharedEngine] playEffect:@"player_joined.mp3"];
            break;
        case kSoundBallOut:
            [[SimpleAudioEngine sharedEngine] playEffect:@"ball_out.mp3"];
            break;
        case kSoundPlayerLeft:
            [[SimpleAudioEngine sharedEngine] playEffect:@"player_left.mp3"];
            break;
        case kSoundBallOnPaddle:
            [[SimpleAudioEngine sharedEngine] playEffect:@"ball_on_paddle.mp3"];
            break;
        case kSoundBallOnWall:
            [[SimpleAudioEngine sharedEngine] playEffect:@"ball_on_wall.mp3"];
            break;
        case kSoundPaddleOnPaddle:
            [[SimpleAudioEngine sharedEngine] playEffect:@"paddle_on_paddle.mp3"];
            break;
        case kSoundPaddleOnWall:
            [[SimpleAudioEngine sharedEngine] playEffect:@"paddle_on_wall.mp3"];
            break;
        default:
            break;
    }
}

#pragma mark - Player management

-(PeerPongPlayer *)addPlayerWithHello:(Hello *)aHello andSocket:(GCDAsyncSocket*)aSocket {
    @synchronized(playerLock) {
        if(![self getPlayerSlotBySocket:aSocket]) {    
            PeerPongPlayer *newPlayer = [[[PeerPongPlayer alloc] init] autorelease];
            newPlayer.socket = aSocket;
            newPlayer.name = aHello.displayName;
            newPlayer.offset = 0.5;
            
            [self.playerQueue addObject:newPlayer];
            return newPlayer;
        } else {
            return nil;
        }
    }
}

-(PeerPongPlayer *)movePlayerFromQueueToSlot:(NSNumber *)aSlot WithSocket:(GCDAsyncSocket *)aSocket {
    @synchronized(playerLock) {
        NSNumber *nextFreeSlot = [self getFreeSlotForPlayer];
        
        if([nextFreeSlot intValue] != [aSlot intValue]) {
            NSLog(@"BUG: Multiple players for one slot received! aSlot: %d, NextFreeSlot: %d", [aSlot intValue], [nextFreeSlot intValue]);
            aSlot = nextFreeSlot;
        }
        
        PeerPongPlayer *player = [self getPlayerSlotBySocket:aSocket];
        player.slot = aSlot;
        [self.players setObject:player forKey:aSlot];
        [self.playerQueue removeObject:player];
        [self playSound:kSoundPlayerJoined];
        [self popMessage:[NSString stringWithFormat:@"Welcome %@!", player.name]];
        
        return player;
    }
}

-(NSNumber *)getFreeSlotForPlayer {
    @synchronized(playerLock) {
        for (int i= 0; i < 4; i++) {
            if([self.players objectForKey:[NSNumber numberWithInt:i]] == nil) {
                return [NSNumber numberWithInt:i];
            }
        }
        return nil;
    }
}

-(void)removePlayerBySocket:(GCDAsyncSocket *)aSocket {
    @synchronized(playerLock) {
        PeerPongPlayer *player = [self getPlayerSlotBySocket:aSocket];
        MainScene *scene = (MainScene *)[[CCDirector sharedDirector] runningScene];
        [self playSound:kSoundPlayerLeft];
        
        if([self.playerQueue containsObject:player]) {
            [self.playerQueue removeObject:player];
            [scene updateQueue];
            [self askFirstPlayerForReadiness];
        } else if([self.players objectForKey:player.slot]) {
            [scene removePaddleForPlayer:player];
            [self.players removeObjectForKey:player.slot];
        }
    }
}

-(PeerPongPlayer *)getPlayerSlotBySocket:(GCDAsyncSocket *)aSocket {
    @synchronized(playerLock) {
        for (PeerPongPlayer *player in [self.players allValues]) {
            if(player.socket == aSocket) {
                return player;
            }
        }
        for (PeerPongPlayer *player in self.playerQueue) {
            if(player.socket == aSocket) {
                return player;
            }
        }
        return nil;
    }
}

-(void)askFirstPlayerForReadiness {
    @synchronized(playerLock) {
        if([self.playerQueue count] != 0) {
            
            PeerPongPlayer *firstPlayer = [self.playerQueue objectAtIndex:0];
            
            
            
            if(!firstPlayer.hasBeenAskedToJoin) {
                AreYouReady *areYouReady = [[[AreYouReady alloc] init] autorelease];
                areYouReady.slot = [self getFreeSlotForPlayer];
                
                NSLog(@"Asking player %@ for Readiness", firstPlayer.name);
                [self.peerPongServer sendObject:areYouReady toClient:firstPlayer.socket];
                firstPlayer.hasBeenAskedToJoin = YES;
            }
            
        }
    }
}

-(void)sendPlayerGoodBye:(PeerPongPlayer *)aPlayer {
    GoodBye *goodBye = [[[GoodBye alloc] init] autorelease];
    [self.peerPongServer sendObject:goodBye toClient:aPlayer.socket];
}

-(void)sendPlayerGameOver:(PeerPongPlayer *)aPlayer {
    GameOver *gameOver = [[[GameOver alloc] init] autorelease];
    [self.peerPongServer sendObject:gameOver toClient:aPlayer.socket];
}

-(void)movePlayerFromSlotToQueue:(PeerPongPlayer *)aPlayer {
    @synchronized(playerLock) {
        if([self.players objectForKey:aPlayer.slot]) {
            MainScene *scene = (MainScene *)[[CCDirector sharedDirector] runningScene];
            [scene removePaddleForPlayer:aPlayer];
            [aPlayer resetLives];
            [self playSound:kSoundPlayerLeft];
            [self.playerQueue addObject:aPlayer];
            [self.players removeObjectForKey:aPlayer.slot];
            [scene updateQueue];
            
            PeerPongPlayer *firstPlayer = [self.playerQueue objectAtIndex:0];
            if(firstPlayer != aPlayer) [self sendPlayerGameOver:aPlayer];
            aPlayer.hasBeenAskedToJoin = NO;
            
            [self askFirstPlayerForReadiness];
        }
    }
}

#pragma mark - Message Pop
-(void)popMessage:(NSString*)aMessage{
    MainScene *scene = (MainScene *)[[CCDirector sharedDirector] runningScene];
    dispatch_async(dispatch_get_main_queue(), ^{
        [scene popMessage:aMessage];
    });
}

#pragma mark - Game callbacks

-(void)ballIsOnPosition:(CGPoint)aPosition {
    if(aPosition.x < 0 || aPosition.y < 0 
       || aPosition.x > self.windowSize.width || aPosition.y > self.windowSize.height) {
        [self ballWentAway:aPosition];
    }
    
}
             
 -(void)ballWentAway:(CGPoint)aPosition {
     int failedPlayerSlot = 0;
     
     if(aPosition.x < 0) {         
         // left player fails (3)
         failedPlayerSlot = 2;
     } else if(aPosition.x > self.windowSize.width) {
         // right player fails (4)
         failedPlayerSlot = 3;
     } else if(aPosition.y > self.windowSize.height) {         
         // top player fails (2)
         failedPlayerSlot = 1;
     }
     
     @synchronized(playerLock) {
         PeerPongPlayer *player = [self.players objectForKey:[NSNumber numberWithInt:failedPlayerSlot]];
         [player loseLife];
         
         LostLife *lostLife = [[[LostLife alloc] init] autorelease];
         lostLife.lifesLeft = [NSNumber numberWithInt:player.livesLeft];
         [self.peerPongServer sendObject:lostLife toClient:player.socket];
         
         MainScene *scene = (MainScene *)[[CCDirector sharedDirector] runningScene];
         if(player.livesLeft < 1) {
             [self movePlayerFromSlotToQueue:player];
             [scene reinitGame];
         } else {
             [self playSound:kSoundBallOut];
             [scene reinitGame];
         }
     }
 }



#pragma mark - Dealloc

- (void)dealloc {
    self.peerPongServer = nil;
    self.players = nil;
    self.playerQueue = nil;
    delete world;
    [playerLock release];
    [super dealloc];
}

@end
